/*
 *  MPEGstream.h - operate with mpeg audio stream
 *
 *  Copyright (C) 2002-2003 Glen Sawyer & Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "MPEGstream.h"

#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "CRC16.h"
#include "mybuffer.h"
#include "ErrorM.h"
#include "bitstream.h"

#define HEADERSIZE 4
#define CRC16_POLYNOMIAL 0x8005


#undef min //absolutely important because of the side effects of the macros
#undef max

template <class X>
inline const X min(const X &a, const X &b)
{  return (a>b)?b:a;
}
template <class X>
inline const X max(const X &a, const X &b)
{  return (a<b)?b:a;
}


// class MPEGstream

MPEGstream::MPEGstream(mybuffer& buf)
 : buf(buf)
 , startup(true)
 , frameOK(true)
 , frame(0)
 , layer(0)
{}

const size_t MPEGstream::bitrates[2][4][16] = { // [version != 1.0][layer_index][bit_rate_index]
{  { 0,     0,     0,     0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0, 0 }
 , { 0, 32000, 40000, 48000,  56000,  64000,  80000,  96000, 112000, 128000, 160000, 192000, 224000, 256000, 320000, 0 }
 , { 0, 32000, 48000, 56000,  64000,  80000,  96000, 112000, 128000, 160000, 192000, 224000, 256000, 320000, 384000, 0 }
 , { 0, 32000, 64000, 96000, 128000, 160000, 192000, 224000, 256000, 288000, 320000, 352000, 384000, 416000, 448000, 0 }
},
{  { 0,     0,     0,     0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      0, 0 }
 , { 0,  8000, 16000, 24000,  32000,  40000,  48000,  56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, 0 }
 , { 0,  8000, 16000, 24000,  32000,  40000,  48000,  56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, 0 }
 , { 0, 32000, 48000, 56000,  64000,  80000,  96000, 112000, 128000, 144000, 160000, 176000, 192000, 224000, 256000, 0 }
}  };

const size_t MPEGstream::frequencies[4][4] = // [version_index][frequency_index]
{  { 11025, 12000,  8000,  0 },
   {     0,     0,     0,  0 },
   { 22050, 24000, 16000,  0 },
   { 44100, 48000, 32000,  0 }
};

const size_t MPEGstream::l3_sideinfo_lens[2][2] = // [!MPEG1][!single_channel]
{  { 17, 32 }
 , { 9,  17 }
};

size_t MPEGstream::frequency() const
{  assert(!startup);
   return frequencies[mpegver >> 3][freq >> 2];
}

void MPEGstream::syncinfo(bool state)
{  if (state == frameOK)
      return;
   frameOK = state;
   *DefEH << (state
    ? Event(Information, 2, "- resync at 0x%lx", buf.total())
    : Event(Warning,     0, "- framesync error at 0x%lx", buf.total()) );
}

bool MPEGstream::skipID3v2()
{
/*
 *  An ID3v2 tag can be detected with the following pattern:
 *    $49 44 33 yy yy xx zz zz zz zz
 *  Where yy is less than $FF, xx is the 'flags' byte and zz is less than
 *  $80.
 */
   if (!buf.fill(10))
      return true; // not enough bytes for an ID3 V2 tag

   unsigned char* wrdpntr = buf.get();

   if ( wrdpntr[0] != 'I' || wrdpntr[1] != 'D' || wrdpntr[2] != '3'
       || wrdpntr[3] == 0xFF || wrdpntr[4] == 0xFF)
      return true; // no ID3 V2 tag

   size_t ID3Size = 10 + (
    (((size_t)wrdpntr[6] << 7 | (size_t)wrdpntr[7]) << 7 | (size_t)wrdpntr[8]) << 7 | wrdpntr[9] );

   return buf.skip(ID3Size); // request at least an MPEG header
}

bool MPEGstream::skipXing()
{  size_t sideinfo_end = HEADERSIZE + l3_sideinfo_lens[(buf[1] & 0x08) != 0x08][(buf[3] & 0xC0) != 0xC0];
   if (!(buf[1] & 0x01)) // CRC
      sideinfo_end += 2;

   if (!buf.fill(sideinfo_end + 4))
      return true; // file too short => no Xing header

   unsigned char* Xingcheck = buf.get() + sideinfo_end;
   if (Xingcheck[0] != 'X' || Xingcheck[1] != 'i' || Xingcheck[2] != 'n' || Xingcheck[3] != 'g')
      return true; // no Xing header

   return buf.skip(framelen()) && frameSearch();
}

bool MPEGstream::firstFrame()
{  layer = buf[1] & 0x06;
   mpegver = buf[1] & 0x18;
   freq = buf[2] & 0x0C;
   mode = buf[3] & 0xC0;
   if ( mpegver == 0x08    // invalid MPEG version
       || layer == 0x00 )  // invalid Layer
      return false;
   startup = false;

   if ((buf[2] >> 4) == 0x00)
      *DefEH << Event(Error, 30, "File is free format (not currently supported)"); // throws!
   // fill arrbytesinframe cache
   slotsize = 1;
   size_t bitbase;
   if (layer == 0x06)
   {  slotsize = 4;
      bitbase = 384;
   } else if (mpegver == 0x18)
      bitbase = 1152;
    else
      bitbase = 576;
   const size_t* ap = bitrates[mpegver != 0x18][layer >> 1];
   size_t freq_cache = frequency();
   if (freq_cache == 0)
      return false; // frequency bad
   for (int i = 0; i < 16; i++)
      arrbytesinframe[i] = ((bitbase * ap[i] / freq_cache) >> 3) & -slotsize; // / 8 bits per byte, rounded to multiples of slotsize
   // got it, first valid frame
   return true;
}

unsigned char* MPEGstream::frameSearch()
{  // ID3 V2 tag ?
   if (startup)
      if (!skipID3v2())
         return NULL;

   unsigned char* wrdpntr;

   while (true)
   {  if (!buf.fill(HEADERSIZE))
         return NULL;
      wrdpntr = buf.get();

      if ((wrdpntr[0] & 0xFF) != 0xFF)    // first 8 bits must be '1'
      {  if (wrdpntr[0] == 'T' && wrdpntr[1] == 'A' && wrdpntr[2] == 'G')
         {  buf.skip(128); // ID3 V1 Tag
            continue; // not an error
         }
         // probably just corrupt data, keep trying
      } else if ( (wrdpntr[1] & 0xE0) == 0xE0 // next 3 bits are also '1'
          && (wrdpntr[2] & 0xF0) != 0xF0  // bad bitrate
          && (layer == 0 || (wrdpntr[1] & 0x06) == layer) ) // not the same layer
      {  if (startup)
         {  // first frame
            if (firstFrame())
               // OK, first valid frame
               break;
         // !startup -- if MPEG version or frequency is different, then probably not correctly synched yet
         } else if ( (wrdpntr[1] & 0x18) == mpegver // version must not change
             && (wrdpntr[2] & 0x0C) == freq  // frequency must not change
            // some Amazon files are currupt, so let's be not that accurate
            // && (wrdpntr[3] & 0xC0) == mode  // mode must not change
             && (wrdpntr[2] & 0xF0) != 0 ) // bitrate is "free format" probably just
            break;                         // corrupt data if we've already found valid frames
      }

      // bitstream error
      syncinfo(false);
      buf.skip(1);
   }
   // got it
   syncinfo(true);
   lastframelen = arrbytesinframe[buf[2] >> 4 /*& 0x0F*/] + ((buf[2] >> 1) & 0x01) * slotsize;
   if (frame == 0 && !skipXing())  // skip xing header
      return NULL;
   ++frame;
   if (!buf.fill(lastframelen)) // Take care of MPEG files where the last frame is incomplete.
      lastframelen = buf.len(); // The FhG encoder is a favorite source of this.
   return wrdpntr;
}

unsigned char* MPEGstream::audio_start() const
{  assert(!startup && frameOK);
   return buf.get() + (!(buf[1] & 0x01) ? 4+2 : 4);
}

unsigned short MPEGstream::calcCRC(unsigned int audio_bits) const
{  CRC16 crc(CRC16_POLYNOMIAL, 0xffff); /* (jo) init crc16 for error_protection */

   crc.Update(buf.get()+2, 16);  // header
   crc.Update(buf.get()+6, audio_bits); // audio data

   return crc.Get();
}

bool MPEGstream::verifyCRC(unsigned int audio_bits) const
{  if (buf[1] & 0x01)
      return true; // no CRC
   unsigned int crc = calcCRC(audio_bits);
   return buf[4] == crc >> 8 && buf[5] == (unsigned char)crc;
}

void MPEGstream::updateCRC(unsigned int audio_bits)
{  if (buf[1] & 0x01)
      return; // no CRC
   unsigned int crc = calcCRC(audio_bits);
   buf[4] = crc >> 8;
   buf[5] = (unsigned char)crc;
}

double MPEGstream::int2factor(int val)
{  assert(!startup);
   return pow(2, val/(layer == 0x02 ? 4. : 3.));
}

double MPEGstream::factor2int(double factor)
{  assert(!startup);
   return (layer == 0x02 ? 4./log(2) : 3./log(2)) * log(factor);
}


// class MPEGgain


/*MPEGgain::MPEGgain(mybuffer& buf, double factor, int gain)
 : MPEGstream(buf)
{  gainfactor[0] = gainfactor[1] = factor;
   gainchange[0] = gainchange[1] = gain;
}

MPEGgain::MPEGgain(mybuffer& buf, double lfact, int left, double rfact, int right)
 : MPEGstream(buf)
{  gainfactor[0] = lfact;
   gainfactor[1] = rfact;
   gainchange[0] = left;
   gainchange[1] = right;
}*/

MPEGgain::MPEGgain(mybuffer& buf, int left, int right)
 : MPEGstream(buf)
{  gainchange[0] = left;
   gainchange[1] = right;
}

int MPEGgain::handlel3gain(unsigned char* data, int shift, int delta)
{  int raw = (data[0] << 8) | data[1];
   int val = (raw >> shift) & 0xFF;
   //printf("%i%3i ", ch, val);
   if (!val) // if val was 0 it is a good idea to leave val=0
      return val;
   if (delta)
      val += delta;
   if (!wrapGain)
   {  if (val < 0)
      {  *DefEH << Event(Warning, 4, "Gain underflow (%i).", val);
         val = 0;
      } else if (val > 255)
      {  *DefEH << Event(Warning, 4, "Gain overflow (%i).", val);
         val = 255;
      }
   } else
      val &= 255;
   raw &= ~(0xFF << shift);
   raw |= val << shift;
   data[0] = (unsigned char)(raw >> 8);
   data[1] = (unsigned char)raw;
   return val;
}

int MPEGgain::layerIIIframe()
{  static const unsigned int CRClen[4] = { 72, 136, 136, 256 }; // number of protected bits

   int sitype = (mpegver == 0x18) | ((mode != 0xC0) << 1);
   unsigned char* sibuf = audio_start();
   // read sideinfo
   int maxgain;
   //printf("-%i-", sitype);
   switch (sitype)
   {default:// mono, lsf
      maxgain = handlel3gain(sibuf+3, 2, gainchange[0]);
      break;
    case 1: // mono
      maxgain = max(handlel3gain(sibuf+4, 1, gainchange[0]), handlel3gain(sibuf+12, 6, gainchange[0]));
      break;
    case 2: // stereo, lsf
      maxgain = max(handlel3gain(sibuf+3, 1, gainchange[0]), handlel3gain(sibuf+11, 2, gainchange[1]));
      break;
    case 3: // stereo
      maxgain = max( max(handlel3gain(sibuf+5, 7, gainchange[0]), handlel3gain(sibuf+12, 4, gainchange[1])),
                     max(handlel3gain(sibuf+19, 1, gainchange[0]), handlel3gain(sibuf+27, 6, gainchange[1])) );
   }

   updateCRC(CRClen[sitype]);

   return maxgain;
}

#include "mpg123.h"
#include "mpglib.h"
#include "layer2.h"
#include "l2tables.h"

void MPEGgain::handlescalefac(bitstream& frm, int delta, int n)
{  while (n--)
   {  int val = 63-frm.peek(6);
      val += delta;
      if (!wrapGain)
      {  if (val < 0)
         {  *DefEH << Event(Warning, 4, "Gain underflow (%i).", val);
            val = 0;
         } else if (val > 63)
         {  *DefEH << Event(Warning, 4, "Gain overflow (%i).", val);
            val = 63;
         }
      } else
         val &= 63;
      frm.set(6, 63-val); // write back
}  }

int MPEGgain::layerIIframe()
{  static const unsigned char translate[3][2][16] = // [sampling_rate][single_channel][bit_rate_index]
    { { { 0,2,2,2,2,2,2,0,0,0,1,1,1,1,1,0 } ,
        { 0,2,2,0,0,0,1,1,1,1,1,1,1,1,1,0 } } ,
      { { 0,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0 } ,
        { 0,2,2,0,0,0,0,0,0,0,0,0,0,0,0,0 } } ,
      { { 0,3,3,3,3,3,3,0,0,0,1,1,1,1,1,0 } ,
        { 0,3,3,0,0,0,1,1,1,1,1,1,1,1,1,0 } } };
   static const struct al_table2* tables[5] = { alloc_0, alloc_1, alloc_2, alloc_3 , alloc_4 };
   static const int sblims[5] = { 27, 30, 8, 12, 30 }; // [table]
   static const size_t CRCbits[2][5] = // [single_channel][table]
    { { 284, 308, 84, 124, 308 } , // I don't know if the lsf modes are correctly set
      { 142, 154, 42,  62, 154 } };

   // audio data segment
   bitstream frm(audio_start());

   // fetch allocation table and subband limits
   int table = mpegver != 0x18 // Version > 1.0 => low frequency
    ? 4
    : translate[freq>>2][mode == 0xC0][buf[2] >> 4];

   int sblim = sblims[table];
   const al_table2* alloc = tables[table];
   int jsbound;
   switch (mode)
   {case 0xC0: // mono
      jsbound = 0;
      break;
    case 0x40: // intensity stereo
      jsbound = (((buf[3] >> 4) & 0x03) +1) << 2; // (mode_extension +1) * 4
      break;
    default:
      jsbound = sblim;
   }
   //printf("at: %i, sblim: %i, jsb: %i\n", table, sblim, jsbound);
   // get the total number of scale factor selection informations
   char chan[32*2][2] = {{0}}; // no more than 2*32 scale factors
   char (*chp)[2] = chan;
   int step;
   for (int i = jsbound; i--; alloc += 1<<step)
   {  step = alloc->bits;
      chp += frm.get(step) != 0; // left
      if (frm.get(step) != 0)
         (*chp++)[0] = 1; // right
   }
   for (int i = sblim-jsbound; i--; alloc += 1<<step)
      chp += frm.get(step = alloc->bits) != 0;

   // get the total number of scalefactors
   static const char mapScfsi[4] = {3,2,1,2}; // map scfsi to the number of scalefactors
   char (*chp2)[2] = chan;
   while (chp2 < chp)
      (*chp2++)[1] = mapScfsi[frm.get(2)];

   // scale!
   for (chp2 = chan; chp2 < chp; chp2++)
      handlescalefac(frm, gainchange[(*chp2)[0]], (*chp2)[1]);

   updateCRC(CRCbits[mode == 0xC0][table]);

   return chp - chan;
}

int MPEGgain::layerIframe()
{  if (mpegver != 0x18)
      return INT_MAX;  // currently only V1.0 supported
   // audio data segment
   bitstream frm(audio_start());
   // determine the total number of subband allocation coefficients
   int nSubBandChannels = 32; // subbands per channel
   switch (mode)
   {case 0x80: // dual channel
    case 0x00: // stereo
      nSubBandChannels <<= 1; // twice as much
      break;
    case 0x40: // intensity stereo
      nSubBandChannels += (((buf[3] >> 4) & 0x03) +1) << 2; // number of stereo subbands = 4 * (ModeExtension+1)
      break;
   }
   // fetch the number of scalefactors := number of Subchannels with allocation != 0
   char chan[32*2]; // no more than 2*32 scale factors
   char* chp = chan;
   while (nSubBandChannels--)
      if (frm.get(4) != 0)
         *chp++ = ~nSubBandChannels & 1; // first loop: 63 -> 0
   // scale!
   char* chp2 = chan;
   while (chp2 < chp)
      handlescalefac(frm, gainchange[*chp2++]);

   updateCRC(mode == 0xC0 ? 128 : 256); // no. of CRCbits for lstf: ???

   return chp - chan;
}

bool MPEGgain::firstFrame()
{  if (!MPEGstream::firstFrame())
      return false;
   if ((mode & 0x40) != 0 && gainchange[0] != gainchange[1])
      *DefEH << Event(Error, 30, "Can't adjust single channel for mono or joint stereo."); // throws!
   switch (layer)
   {default: // error
      return false;
    case 0x02: // layer III
      framehandler = &MPEGgain::layerIIIframe; break;
    case 0x04: // layer II
      framehandler = &MPEGgain::layerIIframe; break;
    case 0x06: // layer I
      framehandler = &MPEGgain::layerIframe; break;
   }
   return true;
}

void MPEGgain::doFrame()
{  (this->*framehandler)();
   // commit changes
   buf.commit(framelen());
}

