/*
 *  MPEGstream.h - operate with mpeg audio stream
 *
 *  Copyright (C) 2002-2004 Glen Sawyer & Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef MPEGstream_h
#define MPEGstream_h

#include "mybuffer.h"
#include "ErrorM.h"
#include "bitstream.h"

#include <stdlib.h>
#include <math.h>
#include <assert.h>


// class MPEGstream - synchronization and metadata of MPEG audio streams
class MPEGstream
{protected:
   mybuffer&      buf; // stream buffer handler
   // state
   bool           startup;
   bool           frameOK;
   int            frame;
   size_t         lastframelen;

   // constant MPEG stream properties (set by firstFrame)
   unsigned char  mpegver; // MPEG version << 3
   unsigned char  freq;    // stream frequency index << 2
   unsigned char  layer;   // mpeg layer index << 1
   unsigned char  mode;    // mode index << 6
   size_t         arrbytesinframe[16]; // framesize cache
   size_t         slotsize;

   // tables
   static const size_t bitrates[2][4][16]; // [version != 1.0][layer_index][bit_rate_index]
   static const size_t frequencies[4][4]; // [version_index][frequency_index]
   static const size_t l3_sideinfo_lens[2][2]; // [!MPEG1][!single_channel]

   // Skip ID3V2 tag at the front of a file (if any)
   // Return false on permanent error.
   bool           skipID3v2();
   // Skip Xing header (if any)
   // Return false on permanent error.
   bool           skipXing();
   // Fetch metadata from first frame and do some initializations.
   virtual bool   firstFrame();

 private:
   void           syncinfo(bool state);
   unsigned short calcCRC(unsigned int audio_bits) const;

 public:
   MPEGstream(mybuffer& buf);
   // force MPEG layer
   void           forceLayer(int l)             { layer = (4-l) << 1; }
   // Search the next (first) MPEG frame in the stream.
   // Return pointer to complete frame on success, otherwise NULL.
   unsigned char* frameSearch();
   // Return the number of bytes in the current frame
   size_t         framelen() const              { assert(frameOK); return lastframelen; }
   // Return the sampling frequency or 0 if invalid
   // do not call this function before frameSearch()
   size_t         frequency() const;
   // get MPEG layer
   int            MPEGlayer() const             { return (layer >> 1) ^ 2; }
   // get number of channels.
   // This function does not work until thie first frame is read.
   int            channels() const              { assert(!startup); return 2 - (mode == 0xC0); }
   // get starting point of audio data
   unsigned char* audio_start() const;
   // verify CRC field (if any)
   bool           verifyCRC(unsigned int audio_bits) const;
   // update CRC field (if any)
   void           updateCRC(unsigned int audio_bits);
   // get current frame no
   int            frameno() const               { return frame; }
   // Convert integer gainchange to scalefactor.
   static double  int2factor(int layer, int val){ return pow(2, val/(layer == 3 ? 4. : 3.)); }
   // This function does not work until thie first frame is read.
   double         int2factor(int val);
   // Convert scalefactor to nearest integer gainchange. The rounding is up to you.
   static double  factor2int(int layer, double factor)
                                                { return (layer == 3 ? 4./log(2) : 3./log(2)) * log(factor); }
   // This function does not work until thie first frame is read.
   double         factor2int(double factor);
};


// class MPEGgain - class to modify the gain of the audio data
class MPEGgain : public MPEGstream
{public: // options
   bool           wrapGain;

 protected:
   // double         gainfactor[2]; // unless we do not know the Layer, we can't combine integer and factor scalings.
   int            gainchange[2];
   // constant MPEG stream properties (set by firstFrame)
   int            (MPEGgain::*framehandler)();

 private:
   int            handlel3gain(unsigned char* data, int shift, int delta);
   void           handlescalefac(bitstream& frm, int delta, int n = 1);

 protected:
   int            layerIIIframe();
   int            layerIIframe();
   int            layerIframe();

   virtual bool   firstFrame();

 public:
   //MPEGgain(mybuffer& buf, double factor, int gain);
   //MPEGgain(mybuffer& buf, double lfact, int left, double rfact, int right);
   MPEGgain(mybuffer& buf, int left, int right);
   // modify gain of the current frame
   void           doFrame();
};


#endif // MPEGstream_h
