/*
 *  filechange.cpp - analyzes mp3 files, determines the perceived volume, modify the MPEG files
 *
 *  Copyright (C) 2002-2004 Glen Sawyer & Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "filechange.h"
#include "filefuncs.h"

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>


/* I tweaked the mpglib library just a bit to spit out the raw
 * decoded double values, instead of rounding them to 16-bit integers.
 * Hence the "mpglibDBL" directory
 */
#include "interface.h"
#include "gain_analysis.h"

#include "MPEGstream.h"
#include "mybuffer.h"
#include "rg_error.h"


bool fileinfo::databaseFormat = false;
bool fileinfo::QuietMode = false;
int fileinfo::ForceLayer = 0;


int fileinfo::passError(int lerrnum, int numStrings, ...)
{
    char * errstr;
    size_t totalStrLen = 0;
    int i;
    va_list marker;

    va_start(marker, numStrings);
    for (i = 0; i < numStrings; i++)
        totalStrLen += strlen(va_arg(marker, const char *));
    va_end(marker);

    errstr = (char*)malloc(totalStrLen + 3);
    errstr[0] = '\0';

    va_start(marker, numStrings);
    for (i = 0; i < numStrings; i++)
        strcat(errstr,va_arg(marker, const char *));
    va_end(marker);

    // error codes less than 0 are mapped to MP3GAIN_UNSPECIFED_ERROR to keep compatibility.
    DoError(errstr, lerrnum < 0 ? MP3GAIN_UNSPECIFED_ERROR : lerrnum);
    free(errstr);
    errstr = NULL;

    return lerrnum;
}


void fileinfo::PrintGainInfo() const
{  const char* type;
   const char* dest;
   if (filename)
   {  type = "Track";
      dest = "";
   } else
   {  type = "Album";
      dest = " for all files";
   }
   int intgainch = intgainchange();
   if (databaseFormat)
   {  fprintf(stdout,"%s\t%d\t%.2f\t%.0f\n", filename, intgainch, dBchange, maxsample);
      fflush(stdout);
   } else {
      fprintf(stdout,"Recommended \"%s\" dB change%s: %.2f\n", type, dest, dBchange);
      if (layer != 0)
         fprintf(stdout,"Recommended \"%s\" MPEG gain change%s: %d\n", type, dest, intgainch);
      fprintf(stdout,"Max PCM sample at current gain: %.0f\n", maxsample);
      if (maxsample * (layer != 0 ? MPEGstream::int2factor(layer, intgainch) : fromdB(dBchange)) > 32767.0)
         fprintf(stdout,"WARNING: some clipping may occur with this gain change!\n");
      fprintf(stdout,"\n");
}  }

// return 0 on success
int fileinfo::Analyze()
{  if (!databaseFormat)
      fprintf(stdout,"%s\n", filename);

   long gFilesize = getSizeOfFile(filename);

   FILE* inf = fopen(filename, "rb");
   if (inf == NULL)
   {  fprintf(stdout, "Can't open %s for reading\n\n", filename);
      return 1;
   }
   // analyze file
   MPSTR mp;
   InitMP3(&mp);
   maxsample = 0;

   Float_t lsamples[1152]; // data buffers
   Float_t rsamples[1152];
   int analysisError = 0;
   try
   {  // initialize buffer manager here
      mybuffer buf(inf);
      MPEGstream mpg(buf);

      bool ok = buf.fill(1);

      if (ok)
      {  ok = mpg.frameSearch();

         /*FILE* dbgf = fopen("out.wav", "wb");
         static const char wavhdr[44] =
         { 0x52, 0x49, 0x46, 0x46, 0x48, 0xB1, 0x29, 0x40, 0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20,
           0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x44, 0xAC, 0x00, 0x00, 0x10, 0xB1, 0x02, 0x00,
           0x04, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61, 0x24, 0xB1, 0x29, 0x40};
         fwrite(wavhdr, 44, 1, dbgf);*/

         if (!ok)
            fprintf(stderr, "Can't find any valid MP3 frames in file %s\n\n", filename);
          else
         {  layer = mpg.MPEGlayer();
            if (!maxAmpOnly) {
               if (ok) {
                  static size_t lastfreq = 0;
                  static int first = 1;
                  if (first)
                  {  lastfreq = mpg.frequency();
                     InitGainAnalysis(lastfreq);
                     analysisError = 0;
                     first = 0;
                  } else if (mpg.frequency() != lastfreq)
                  {  lastfreq = mpg.frequency();
                     ResetSampleFrequency(lastfreq);
                  }
               }
            } else
               analysisError = 0;

            while (ok)
            {  unsigned char* curframe = buf.get();

               int bytesinframe = mpg.framelen();
               assert(bytesinframe);
               int nchan = mpg.channels();

               lSamp = lsamples;
               rSamp = rsamples;
               maxSamp = &maxsample;
               procSamp = 0;
               int nprocsamp;
               if (decodeMP3(&mp,curframe,bytesinframe,&nprocsamp) == MP3_OK)
               {  /*float* l = lsamples;
                  float* r = rsamples;
                  //printf("Size: %i\n", procSamp);
                  for (int i = 0; i < procSamp/nchan; ++i)
                  {  short smp[2] = {*l++, *r++};
                     fwrite(smp, 4, 1, dbgf);
                  }*/
                  //printf("bulb %f\n", *lSamp);
                  if (!maxAmpOnly)
                  {  if (AnalyzeSamples(lsamples,rsamples,procSamp/nchan,nchan) == GAIN_ANALYSIS_ERROR)
                     {  fprintf(stderr,"Error analyzing further samples (max time reached)          \n");
                        analysisError = true;
                        ok = 0;
                     }
                  }
               }

               if (!analysisError)
               {  buf.commit(bytesinframe);
                  ok = mpg.frameSearch();
               }

               if (!QuietMode)
               {  if (mpg.frameno() % 200 == 0)
                  {  // fprintf(stderr,"frame %d\r",frame);
                     fprintf(stderr,"\r%zd of %ld bytes analyzed\r", buf.total(), gFilesize);
                     fflush(stderr);
                  }
               }
            }
            if (!QuietMode)
               fprintf(stderr,"                                       \r");
         }
      }
   } catch (const Event& e)
   {  fprintf(stderr, "%s: %s\n", filename, e.Message.c_str());
   }

   ExitMP3(&mp);
   fflush(stderr);
   fflush(stdout);
   fclose(inf);

   fileOK = !analysisError;
   return analysisError;
}


bool filechange::ignoreClipWarning = false;
bool filechange::InteractiveMode = false;
const char* filechange::defInput = "";
bool filechange::autoNoClip = false;
bool filechange::wrapGain = false;
bool filechange::UsingTemp = false;
bool filechange::saveTime = false;


int filechange::CommitGainChange() const
{  const char* name = filename == NULL
    ? "files"
    : filename;
   fflush(stdout);
   int intgainch = intgainchange();
   bool interactive = InteractiveMode; // local copy
   double clip = maxsample * MPEGstream::int2factor
   (layer, intgainch) / 32767.0;
   if (clip > 1 && (intgainch > 0 || autoNoClip))
   {  fprintf(stderr,"WARNING: %s may clip with gain change %d\n", name, intgainch);
      if (!ignoreClipWarning)
         interactive = true;
   }
   if (!interactive && intgainch == 0)
   {  fprintf(stdout,"No changes to %s are necessary\n", name);
      return 0;
   }
   // calculate automatic value
   int aval = intgainch;
   if ((intgainch > 0 || autoNoClip) && clip > 1)
   {  // decrease the automatic value to avoid clipping
      aval -= (int)ceil(MPEGstream::factor2int(layer, clip));
      if (aval < 0 && !autoNoClip)
         aval = 0;
      fprintf(stderr,"INFO: using auto-clipped mp3 gain change of %d (instead of %d) for %s.\n", aval, intgainch, name);
   }
   do
   {  char buf[10] = {0};
      if (interactive)
      {  fprintf(stderr, "Make change? [y(%d)|n(0)|a(%d)|<value>]:", intgainch, aval);
         if (*defInput)
         fprintf(stderr, " (default=%s)", defInput);
         fflush(stderr);
         scanf("%*[ \t]");
         scanf("%9[^ \t\n]", buf);
         scanf("%*[ \t]");
      }
      if (buf[0] == 0)
      {  if (defInput[0] == 0)
            return intgainch; // like 'y'
         strncpy(buf, defInput, 9);
         puts(buf);
      }
      if (buf[1] == 0)
      switch (toupper(buf[0]))
      {case 'Y':
         return intgainch;
       case 'N':
         return 0;
       case 'A':
         return aval;
      }
      int nval;
      size_t n;
      if (sscanf(buf, "%i%n", &nval, (int*)&n) == 1 && n == strlen(buf) && abs(nval) < 60)
         return nval;
      fprintf(stderr, "Bad Input: %s\n", buf);
   } while (interactive);
   exit(32); // permanent error in defInput
}

int filechange::ChangeGain(int leftgainchange, int rightgainchange)
{  if (leftgainchange == 0 && rightgainchange == 0)
      return 0; // NOP

   char *outfilename = NULL;

   if (saveTime)
      fileTime(filename, storeTime);

   size_t gFilesize = getSizeOfFile(filename);

   FILE* inf = NULL;
   FILE* outf = NULL;

   if (UsingTemp)
   {  fflush(stderr);
      fflush(stdout);
      outfilename = (char*)malloc(strlen(filename)+5);
      strcpy(outfilename,filename);
      strcat(outfilename,".tmp");

      inf = fopen(filename,"r+b");

      if (inf != NULL)
      {  outf = fopen(outfilename, "wb");

         if (outf == NULL)
         {  fclose(inf);
            inf = NULL;
            return passError(M3G_ERR_CANT_MAKE_TMP, 3,
               "\nCan't open ", outfilename, " for temp writing\n");
         }
      }
   } else
      inf = fopen(filename,"r+b");

   if (inf == NULL)
   {  if (UsingTemp && (outf != NULL))
         fclose(outf);
      return passError(M3G_ERR_CANT_MODIFY_FILE, 3,
           "\nCan't open ", filename, " for modifying\n");
   }

   // files open. begin to work
   mybuffer* bp = UsingTemp   // initialize buffer manager
    ? new mybuffer(inf, outf) // not in-place
    : new mywbuffer(inf);     // in-place

   MPEGgain mpg(*bp, leftgainchange, rightgainchange);
   if (fileinfo::ForceLayer)
      mpg.forceLayer(fileinfo::ForceLayer);

   bool ok = bp->fill(1);

   try
   {
      if (ok)
      {  ok = mpg.frameSearch();

         if (!ok)
         {  passError( MP3GAIN_UNSPECIFED_ERROR, 3,
                       "Can't find any valid MPEG frames in file ", filename, "\n\n");
         }

   #ifdef asWIN32DLL
         while (ok && (!blnCancel)) {
   #else
         while (ok) {
   #endif
            if (ok)
            {  // gain_change here !
               mpg.doFrame();

               if (!fileinfo::QuietMode)
               {
                  if (mpg.frameno() % 200 == 0) {
   #ifndef asWIN32DLL
                     fprintf(stderr,"\r %2zd%% of %zd bytes written\r", bp->total() * 100 / gFilesize, gFilesize);
                     fflush(stderr);
   #else
                     // report % back to calling app
                     if (sendpercentdone(bp->total() * 100 / gFilesize, gFilesize) != 0)
                        return; // ok
   #endif
                  }
               }
               // next frame
               ok = mpg.frameSearch();
            }
         }
      }
   } catch (const Event& e)
   {  passError( MP3GAIN_FILEFORMAT_NOTSUPPORTED, e.ID, filename, ": ", e.Message.c_str(), "\n");
   }

#ifdef asWIN32DLL
   if (blnCancel) { //need to clean up as best as possible
      fclose(inf);
      const char* addmsg = "";
      if (UsingTemp) {
         fclose(outf);
         deleteFile(outfilename);
         free(outfilename);
      } else
         addmsg = "\nFile is probably corrupted now.";
      if (saveTime)
         fileTime(filename, setStoredTime);
      return passError(MP3GAIN_CANCELLED, 3, "Cancelled processing of ", filename, addmsg);
   }
#endif

   if (!fileinfo::QuietMode) {
#ifndef asWIN32DLL
      fprintf(stderr,"                                                   \r");
#else
      // report DONE (100%) message back to calling app
      sendpercentdone(100, gFilesize);
#endif
   }
   fflush(stderr);
   fflush(stdout);

   if (UsingTemp)
   {  // copy remaining part of file
      while (bp->skip(1))
         bp->commit();
      fflush(outf);
      // verify
      long outlength, inlength; // size checker when using Temp files
#ifdef WIN32
      outlength = filelength(fileno(outf));
      inlength = filelength(fileno(inf));
#else
      fseek(outf, 0, SEEK_END);
      fseek(inf, 0, SEEK_END);
      outlength=ftell(outf);
      inlength =ftell(inf);
#endif
      // keep some platform specific file attributes, details in filefuncs.*
      fileCopyAttributes(inf, outf);

      fclose(outf);
      fclose(inf);
      inf = NULL;

      if (outlength != inlength)
      {  deleteFile(outfilename);
         return passError( M3G_ERR_NOT_ENOUGH_TMP_SPACE, 3,
                "Not enough temp space on disk to modify ", filename,
                "\nEither free some space, or do not use \"temp file\" option\n");
      }
      if (deleteFile(filename) == -1)
         return passError( M3G_ERR_CANT_MODIFY_FILE, 3,
                 "Can't open ", filename, " for modifying\n");
      if (moveFile(outfilename, filename) == -1)
         return passError( M3G_ERR_RENAME_TMP, 9,
                    "Problem re-naming ", outfilename, " to ", filename,
                    "\nThe mp3 was correctly modified, but you will need to re-name ",
                    outfilename, " to ", filename,
                    " yourself.\n");
      free(outfilename);
   } else
   {  // write delaed
      bp->flush();
      fclose(inf);
      inf = NULL;
   }
   delete bp; // buffer no longer needed

   if (saveTime)
      fileTime(filename, setStoredTime);

   return 0;
}

