/*
 *  mybuffer.cpp - read only and read/write I/O buffer manager class
 *
 *  Copyright (C) 2003 Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "mybuffer.h"

#include <string.h>


// mybuffer class - buffer manager
mybuffer::mybuffer(FILE* inf, FILE* outf)
{  InF = inf;
   OutF = outf;
   Len = 0;
   Current = Buffer;
   Bad = false;
   FilePos = 0;
}

void mybuffer::Refresh()
{  size_t inc = Current - Buffer;
   // write back
   if (OutF)
      fwrite(Buffer, inc, 1, OutF); // errors are checked later by the file size
      //if (fwrite(Buffer, inc, 1, Outf) != 1)
      //   throw "Failed to write output stream.";
   FilePos += inc;
   // keep remaining data
   memmove(Buffer, Current, Len);
   Current = Buffer;
   size_t len = fread(Current + Len, 1, sizeof Buffer - Len, InF);
   if (len == 0)
      Bad = true;
   Len += len;
}

bool mybuffer::fill(size_t minreq)
{  assert(minreq <= sizeof Buffer);
   if (Len < minreq && !Bad)
      Refresh();
   return Len >= minreq;
}

void mybuffer::commit(size_t len)
{  assert(len <= Len);
   Current += len;
   Len -= len;
}

bool mybuffer::skip(size_t len)
{  while (len > Len)
   {  len -= Len;
      mybuffer::commit(Len);
      mybuffer::fill(1);
      if (Len == 0) // error
         return false; // skip exceeded the file size
   }
   mybuffer::commit(len);
   return true;
}

bool mybuffer::flush()
{  return true;
}


// mywbuffer class - buffer manager with intellegent write back cache in case of inplace operation
mywbuffer::mywbuffer(FILE* inf)
 : mybuffer(inf)
{  LastReq = 0;
   Head = NULL;
   Tail = &Head;
}

mywbuffer::~mywbuffer()
{  // free mem
   while (Head)
   {  wbuffer* wp = Head;
      Head = wp->Link;
      free(wp);
}  }

void mywbuffer::FillDiffBuffer(size_t len)
{  if (len > LastReq)
   {  memcpy(DiffBuffer, Current, len);
      LastReq = len;
}  }

void mywbuffer::AddWBuffer(size_t pos, unsigned char* data, size_t len)
{  wbuffer* wp = *Tail = (wbuffer*)malloc(sizeof(wbuffer) + len);
   if (wp == NULL)
   {  flush();
      fseek(InF, FilePos + Len, SEEK_SET);
      wp = *Tail = (wbuffer*)malloc(sizeof(wbuffer) + len); // retry
      assert(wp != NULL); // can't help it
   }
   wp->Link = NULL;
   Tail = &wp->Link;
   wp->FilePos = pos;
   wp->Len = len;
   memcpy(wp->Data, data, len);
}

bool mywbuffer::fill(size_t minreq)
{  //printf("fill: %u %u\n", minreq, LastReq);
   assert(minreq <= sizeof DiffBuffer);
   if (mybuffer::fill(minreq))
   {  FillDiffBuffer(minreq);
      return true;
   } else
   {  FillDiffBuffer(Len);
      return false;
}  }

void mywbuffer::commit(size_t len)
{  //printf("commit: %u %u\n", len, LastReq);
   assert(len <= LastReq);
   // compare and create write buffer
   unsigned char* sp = Current;
   unsigned char* dp = DiffBuffer;
   size_t l = len;
   while (l-- && *sp++ == *dp++); // find first different
   if (l != (size_t)-1) // any difference ?
   {  --sp;
      unsigned char* sp2 = Current + len;
      dp = DiffBuffer + len;
      while (*--sp2 == *--dp); // find last different
      AddWBuffer(FilePos + (sp - Buffer), sp, sp2 - sp +1);
   }
   LastReq = 0;
   return mybuffer::commit(len);
}

bool mywbuffer::skip(size_t len)
{  //printf("skip: %u %u\n", len, LastReq);
   LastReq = 0;
   return mybuffer::skip(len);
}

bool mywbuffer::flush()
{  //printf("flush: %u\n", LastReq);
   while (Head)
   {  wbuffer* wp = Head;
      if ( fseek(InF, wp->FilePos, SEEK_SET)
          || fwrite(wp->Data, wp->Len, 1, InF) != 1 )
         return false;
      // next buffer
      Head = wp->Link;
      free(wp);
   }
   Tail = &Head;
   return true;
}

