/*
 *  mybuffer.h - read only and read/write I/O buffer manager class
 *
 *  Copyright (C) 2003 Marcel Mueller
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef mybuffer_h
#define mybuffer_h

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>


#define BUFFERSIZE 100000
#define MAXREQSIZE 5000


/* mybuffer class - buffer manager
*/
class mybuffer
{protected:
   FILE*          InF;
   FILE*          OutF;
   unsigned char  Buffer[BUFFERSIZE];
   unsigned char* Current;
   size_t         Len;
   bool           Bad;
   size_t         FilePos;

   virtual void   Refresh();
 public:
   // construct mybuffer. read from inf and optionally write to outf
   // inf and outf may be transient streams.
                  mybuffer(FILE* inf, FILE* outf = NULL);

   // *** buffer management functions

   // Fill buffer with at least minreq bytes.
   // return true if at least minreq bytes read.
   // fill uses caching.
   // minreq must not exceed the buffer size.
   virtual bool   fill(size_t minreq);
   // commit the next num bytes in the buffer.
   // num should not exceed minreq from the last fill call.
   // num must not exceed len().
   // Any subsequent call to fill will conut the number of bytes from the last commited or skipped byte.
   // That is an no-op unless num is larger.
   virtual void   commit(size_t num);
   // commit the whole buffer
   void           commit()                   { commit(Len); }
   // Skip the next num bytes in the source stream.
   // num is not limited by anything but the number of bytes in the source stream.
   // Return true if the source stream has enough data.
   // When outf is not null, skip does not skip the copying from inf to outf.
   virtual bool   skip(size_t num);          // unlike commit skip may exceed Len
   // Flush any data in the write back cache and return true on success.
   // Since mybuffer has no write back cache flush always returns true.
   // Flush will not thansfer any remaining bytes in inf.
   virtual bool   flush();

   // *** buffer access functions.

   // Results are invalidated by *all* buffer management functions.
   // Get current buffer. The buffer is mutable.
   unsigned char* get() const                { return Current; }
   // Get current buffer length.
   // This length might exceed the number of requested bytes significantly.
   size_t         len() const                { return Len; }
   // Total number of bytes commited since the class is constructed.
   size_t         total() const              { return FilePos + (Current - Buffer); }
   // Direct access to the buffer content. Mainly for convenience.
   unsigned char& operator[](size_t i)       { assert(i < Len); return Current[i]; }
};


/* mywbuffer class - buffer manager with intellegent write back cache designed for inplace operations.
*/
class mywbuffer : public mybuffer
{protected:
   struct wbuffer
   {  struct wbuffer* Link;
      size_t      FilePos;
      size_t      Len;
      unsigned char Data[]; // followed by data
   };

   unsigned char  DiffBuffer[MAXREQSIZE];
   size_t         LastReq;
   wbuffer*       Head;
   wbuffer**      Tail;

   virtual void   FillDiffBuffer(size_t len);
   virtual void   AddWBuffer(size_t pos, unsigned char* data, size_t len);
 public:
   // construct mywbuffer. read from inf and write to inf.
   // inf must not be a transient stream.
                  mywbuffer(FILE* inf);
   // destructor. Does *not* fluch the cache.
   virtual        ~mywbuffer();
   // See mybuffer::fill
   // Dou to the nature of the cache minreq must not exceed MAXREQSIZE
   virtual bool   fill(size_t minreq);
   // commit the next num bytes in the buffer.
   // Unlike mybuffer::commit num *must* not exceed minreq from the last fill call.
   virtual void   commit(size_t num);
   using          mybuffer::commit;
   // See mybuffer::skip
   virtual bool   skip(size_t num);          // unlike commit skip may exceed Len
   // flush the write back cache.
   virtual bool   flush();
};


#endif // mybuffer_h
